#!/usr/bin/env python
# coding: utf-8

"""
Tools for the generation and analysis of dislocation distributions.

A dislocation is characterized by a position and a Burgers vector. A
distribution is a set of dislocations in a material. The dislocations
of a distribution are generated randomly according to a model. A sample
is a set of distributions generated with the same model.

A distribution can be exported as a map showing the location of its
dislocations. It can also be exported in a standardized form for input
to an X-ray diffraction simulation program. Finally it is possible to
make a spatial analysis of a model by means of statistical functions
on a sample of distributions.

In this package, Burgers vectors are represented only by their sense
(positive: +1, or negative: -1). Their direction is only determined in
the header of the exported input data files. It is parallel to the
dislocation line for the 'screw' type or perpendicular for the 'edge'
type. The superposition principle simplifies the analysis by allowing
to study only one direction at a time.

To obtain relevant statistical results, the spatial analysis must be
averaged over a large number of distributions. To reduce the
computation time it is possible to parallelize the calculations. The
MPI library can be used for this purpose. A sample of distributions is
generated and analyzed on each core, then the results are averaged
over the cores.
"""

__author__ = "Dunstan Becht"
__version__ = "1.1.2"

import os
import sys
import warnings
from typing import Union, Optional, Any

if sys.version_info[0]>=3 and sys.version_info[1]>=9:
    from collections.abc import Callable
else:
    from typing import Callable

if sys.version_info[0]>=3 and sys.version_info[1]>=8:
    from typing import get_args, get_origin
else:
    get_args = lambda x: x.__args__
    get_origin = lambda x: x.__origin__ if hasattr(x, '__origin__') else x

import numpy as np
from beartype import beartype

# scalar and vectors
Scalar = Union[int, np.integer, float, np.floating]
Vector = np.ndarray # shape: (n,)

# sets
ScalarList = np.ndarray # shape: (...,)
ScalarListList = np.ndarray # shape: (..., ...)
VectorList = np.ndarray # shape: (..., n)

# generic analysis function output
AnalysisOutput = Union[tuple, Scalar, np.ndarray]

# edge correction functions
CorrectionFunction = Callable[[Vector, ScalarList, ScalarList], ScalarList]

# model generation functions
GenerationFunction = Callable[[str, Scalar, Scalar, dict], tuple]

def getkwa(
    key: str,
    kwa: dict,
    typ: type,
    dft: Any,
) -> Any:
    """
    Retrieve the keyword argument and check its type.

    Input:
        key (str): keyword argument name
        kwa (dict): keyword arguments dictionary
        typ (type): keyword argument type
        dft (Any): keyword argument default value

    Output:
        val (Any): value of the keyword argument
    """
    val = kwa.pop(key, dft)
    if get_origin(typ) == Union:
        typ = get_args(typ)
    if not isinstance(val, typ):
        msg = ("keyword argument '"+key+"' must be of type "+str(typ)
            + " but "+str(type(val))+" given")
        raise TypeError(msg)
    return val

def endkwa(
    kwa: dict,
) -> None:
    """
    Check that no bad keyword arguments have been passed.

    Input:
        kwa (dict): keyword arguments dictionary
    """
    if len(kwa) > 0:
        msg = ("unexpected keyword argument(s): "
            + ", ".join([repr(key) for key in kwa]))
        raise TypeError(msg)



#!/usr/bin/env python
# coding: utf-8

"""
Tools for spatial analysis of dislocation distributions.
"""

import matplotlib.pyplot as plt
import scipy.special
from . import *
from . import sets
from . import geometries
from . import boundaries

@beartype
def N(
    a: Vector,
    B: VectorList,
    r2: ScalarList,
) -> ScalarList:
    """
    Return the number of points of B in the neighborhoods of a.

    The function does not use the notion of dislocation. It allows only
    a spatial analysis of a distribution of points in space. If a point
    of B coincides with the point a then it is not counted.

    Input:
        a (Vector): point around which the neighborhoods are formed
        B (VectorList): points observed and potentially counted
        r2 (ScalarList): squared neighborhood radii in ascending order

    Output:
        n (ScalarList): number of points for each radius value

    Input example:
        a = np.array([a_x, a_y])
        B = np.array([[b_0_x, b_0_y], [b_1_x, b_1_y], ...])
        r2 = np.array([r_0^2, r_1^2, r_2^2, ...])

    Output example:
        n = np.array([N(a, B, r_0), N(a, B, r_1), N(a, B, r_2), ...])

    Complexity:
        O( len(B)*log(len(B)) + len(r2) )
    """
    d2 = np.sum(np.square(np.subtract(B, a)), axis=1) # squared distances to a
    d2 = np.sort(d2[(d2<=r2[-1]) & (d2>0)]) # sorted and filtered sqrd. dist.
    j, dn = 0, np.zeros(len(r2), dtype=np.int64)
    for i in range(len(d2)):
        while d2[i]>r2[j]:
            j += 1
        dn[j] += 1
    return np.cumsum(dn)

@beartype
def M(
    A: VectorList,
    B: VectorList,
    w: CorrectionFunction,
    r: ScalarList,
    r2: ScalarList,
) -> ScalarList:
    """
    Return the value of N(a, B, r2) averaged over the points a in A.

    The function does not use the notion of dislocation. It allows only
    a spatial analysis of a distribution of points in space.

    Input:
        A (VectorList): points of the centers of the neighborhoods
        B (VectorList): points observed and potentially counted
        w (CorrectionFunction): weighting function for edge correction
        r (ScalarList): neighborhood radii in ascending order
        r2 (ScalarList): squared neighborhood radii in ascending order

    Output:
        m (ScalarList): average number of points for each radius value

    Input example:
        A = np.array([[a_0_x, a_0_y], [a_1_x, a_1_y], ...])
        B = np.array([[b_0_x, b_0_y], [b_1_x, b_1_y], ...])
        w = lambda a, r, r2: np.ones(len(r))
        r = np.array([r_0, r_1, r_2, ...])
        r2 = np.array([r_0^2, r_1^2, r_2^2, ...])

    Output example:
        m = np.array([M(A, B, r_0), M(A, B, r_1), M(A, B, r_2)])

    Complexity:
        O( len(A) * (complexity_of(w)+complexity_of(N)) )
    """
    sumN = np.zeros(len(r))
    sumw = np.zeros(len(r))
    for i in range(len(A)):
        sumN += N(A[i], B, r2)
        sumw += w(A[i], r, r2)
    return sumN/sumw

@beartype
def MMMM_cp_cm(
    d: sets.Distribution,
    r: ScalarList,
    r2: ScalarList,
    ec: str,
) -> tuple:
    """
    Return M++, M-+, M+-, M-- and the number of + and - dislocations.

    M++ is the average number of dislocations with positive Burgers
    vector sense around dislocations with positive sense. The
    dislocation in the center of the neighborhood is not counted.

    M-+ is the average number of dislocations with positive Burgers
    vector sense around dislocations with negative sense.

    M+- is the average number of dislocations with negative Burgers
    vector sense around dislocations with positive sense.

    M-- is the average number of dislocations with negative Burgers
    vector sense around dislocations with negative sense. The
    dislocation in the center of the neighborhood is not counted.

    When the parameter ec equals 'NEC' no edge correction is performed.
    When it equals 'WOA' the results are weighted by edge correction
    coefficients. When it equals 'PBC' or 'GBB', corresponding boundary
    conditions are applied with a rank of replication determined by the
    maximum value of the neighborhood radius.

    Input:
        d (Distribution): distribution of dislocations to analyze
        r (ScalarList): neighborhood radii in ascending order [nm]
        r2 (ScalarList): squared neighborhood radii [nm^2]
        ec (str): edge consideration

    Output:
        MMMM (ScalarListList): stacked M++, M-+, M+-, M-- values [1]
        cp (Scalar): number of dislocations with sense + [1]
        cm (Scalar): number of dislocations with sense - [1]

    Output example:
        MMMM = np.array([
            [M++(r_0), M++(r_1), M++(r_2), ...],
            [M-+(r_0), M-+(r_1), M-+(r_2), ...],
            [M+-(r_0), M+-(r_1), M+-(r_2), ...],
            [M--(r_0), M--(r_1), M--(r_2), ...],
        ])
        cp = 2
        cm = 3

    Complexity:
        O( complexity_of(M) )
    """
    # masks
    mskroi = geometries.mask(d.g, d.s, d.p) # the dislocation is in the ROI
    mskbsp = d.b>0 # the dislocation has a Burgers vector sense +
    mskbsm = d.b<0 # the dislocation has a Burgers vector sense -
    # neighbourhood centers
    Pp1 = d.p[mskbsp & mskroi] # positions is the ROI with sense +
    Pm1 = d.p[mskbsm & mskroi] # positions is the ROI with sense -
    # observed position
    if ec == 'NEC': # no edge correction
        Pp2 = d.p[mskbsp] # positions with sense +
        Pm2 = d.p[mskbsm] # positions with sense +
    elif ec == 'WOA': # weighting by overlapping area
        Pp2 = Pp1 # observed positions with sense +
        Pm2 = Pm1 # observed positions with sense -
    elif ec=='PBC' or ec=='GBB': # apply boundary conditions
        rep = int(np.ceil(np.max(r)/d.s))
        cp, cb = d.conditions(ec+str(rep)) # outer dislocations
        P2 = np.concatenate((d.p, cp)) # observed positions
        B2 = np.concatenate((d.b, cb)) # observed Burgers vector senses
        Pp2 = P2[B2>0] # observed positions with sense +
        Pm2 = P2[B2<0] # observed positions with sense -
        if ec=='PBC' and r[-1]>=d.s:
            msg = ("discontinuities of M++ and M-- when using PBC "
                + "(This is due to the periodic presence of the same "
                + "dislocation at the same place in the replicated "
                + "regions. The dislocation is suddenly counted "
                + "multiple times at radius values corresponding to "
                + "multiples of d.s, sqrt(2)*d.s etc...)")
            warnings.warn(msg, Warning)
    else:
        raise ValueError("invalid edge consideration: "+str(ec))
    # weighting
    if ec == 'WOA': # no edge correction
        w = d.w # weighting function
    else:
        w = lambda a, r, r2: 1 # weighting function
    # calculate
    Mpp = M(Pp1, Pp2, w, r, r2) # M++
    Mmp = M(Pm1, Pp2, w, r, r2) # M-+
    Mpm = M(Pp1, Pm2, w, r, r2) # M+-
    Mmm = M(Pm1, Pm2, w, r, r2) # M--
    MMMM = np.stack((Mpp, Mmp, Mpm, Mmm)) # stacked M++, M-+, M+-, M--
    cp = len(Pp1) # number of dislocations with sense +
    cm = len(Pm1) # number of dislocations with sense -
    return MMMM, cp, cm

@beartype
def KKKK_dp_dm(
    MMMM: ScalarListList,
    cp: Scalar,
    cm: Scalar,
    v: Scalar,
) -> tuple:
    """
    Return K++, K-+, K+-, K-- and densities of + and - dislocations.

    Implementation of the Ripley’s K function.

    This function is incremental. It uses pre-calculated MMMM values
    to avoid unnecessary repetition of calculations common to spatial
    analysis functions

    Input:
        MMMM (ScalarListList): stacked M++, M-+, M+-, M-- values [1]
        cp (Scalar): number of dislocations with sense + [1]
        cm (Scalar): number of dislocations with sense - [1]
        v (Scalar): n-volume of the region of interest [nm^n]

    Output:
        KKKK (ScalarListList): stacked K++, K-+, K+-, K-- values [nm^n]
        dp (Scalar): density of dislocations with sense + [nm^-n]
        dm (Scalar): density of dislocations with sense - [nm^-n]

    Output example:
        KKKK = np.array([
            [K++(r_0), K++(r_1), K++(r_2), ...],
            [K-+(r_0), K-+(r_1), K-+(r_2), ...],
            [K+-(r_0), K+-(r_1), K+-(r_2), ...],
            [K--(r_0), K--(r_1), K--(r_2), ...],
        ])
        dp = 0.002
        dp = 0.003

    Complexity:
        O( len(MMMM) )
    """
    dp, dm = cp/v, cm/v # densities of + and - dislocations
    KKKK = np.concatenate((MMMM[0:2]/dp, MMMM[2:4]/dm))
    return KKKK, dp, dm

@beartype
def gggg_dV_dKKKK(
    KKKK: ScalarListList,
    dr: ScalarList,
    r: ScalarList,
    n: int,
) -> tuple:
    """
    Return g++, g-+, g+-, g-- and differentials of V and KKKK.

    Implementation of the pair correlation function.

    This function is incremental. It uses pre-calculated KKKK values
    to avoid unnecessary repetition of calculations common to spatial
    analysis functions

    Input:
        KKKK (ScalarListList): stacked K++, K-+, K+-, K-- values [nm^n]
        dr (ScalarList): differentials of the radius [nm]
        r (ScalarList): neighborhood radii in ascending order [nm]
        n (int): dimension of the space of the region of interest

    Ouput:
        gggg (ScalarListList): stacked g++, g-+, g+-, g-- values [1]
        dV (ScalarList): differentials of the neighborhood [nm^n]
        dKKKK (ScalarList): differentials of KKKK [nm^n]

    Output example:
        gggg = np.array([
            [g++(r_0), g++(r_1), g++(r_2), ...],
            [g-+(r_0), g-+(r_1), g-+(r_2), ...],
            [g+-(r_0), g+-(r_1), g+-(r_2), ...],
            [g--(r_0), g--(r_1), g--(r_2), ...],
        ])
        dV = np.array([pi*r_0*dr_0, pi*r_1*dr_1, ...])
        dKKKK = np.array([
            [K++(r_1)-K++(r_0), (K++(r_2)-K++(r_0))/2, ...],
            [K-+(r_1)-K-+(r_0), (K-+(r_2)-K-+(r_0))/2, ...],
            [K+-(r_1)-K+-(r_0), (K+-(r_2)-K+-(r_0))/2, ...],
            [K--(r_1)-K--(r_0), (K--(r_2)-K--(r_0))/2, ...],
        ])

    Complexity:
        O( len(r) )
    """
    dV = (np.pi**(n/2)/scipy.special.gamma(n/2+1))*n*r**(n-1)*dr # diff. of V
    dV[0] = np.nan # mask first value because of zero or wrong gradient
    dV[-1] = np.nan # mask last value because of wrong gradient
    dKKKK = np.gradient(KKKK, axis=1) # differential of KKKK
    gggg = dKKKK/dV
    return gggg, dV, dKKKK

@beartype
def GaGs_dMMMM(
    MMMM: ScalarListList,
    dr: ScalarList,
    cp: Scalar,
    cm: Scalar,
) -> tuple:
    """
    Return Ga and Gs.

    Implementation of the antisymmetrical and symmetrical functions.

    This function is incremental. It uses pre-calculated MMMM values
    to avoid unnecessary repetition of calculations common to spatial
    analysis functions

    Input:
        MMMM (ScalarListList): stacked M++, M-+, M+-, M-- values [1]
        dr (ScalarList): differentials of the radius [nm]
        cp (Scalar): number of dislocations with sense + [1]
        cm (Scalar): number of dislocations with sense - [1]

    Output:
        GaGs (ScalarListList): stacked values of Ga and Gs [nm^-1]
        dMMMM (ScalarList): differentials of MMMM [1]

    Output example:
        GaGs = np.array([
            [Ga(r_0), Ga(r_1), Ga(r_2), ...],
            [Gs(r_0), Gs(r_1), Gs(r_2), ...],
        ])
        dMMMM = np.array([
            [M++(r_1)-M++(r_0), (M++(r_2)-M++(r_0))/2, ...],
            [M-+(r_1)-M-+(r_0), (M-+(r_2)-M-+(r_0))/2, ...],
            [M+-(r_1)-M+-(r_0), (M+-(r_2)-M+-(r_0))/2, ...],
            [M--(r_1)-M--(r_0), (M--(r_2)-M--(r_0))/2, ...],
        ])

    Complexity:
        O( len(MMMM) )
    """
    dMMMM = np.gradient(MMMM, axis=1) # differentials of MMMM
    dMppdr, dMmpdr, dMpmdr, dMmmdr = dMMMM/dr # derivatives of MMMM
    Ga = cp*(dMppdr-dMpmdr) + cm*(dMmmdr-dMmpdr) # antisymmetrical function
    Gs = cp*(dMppdr+dMpmdr) + cm*(dMmmdr+dMmpdr) # symmetrical function
    return np.stack((Ga, Gs)), dMMMM

@beartype
def calculate(
    q: Union[list, tuple],
    o: Union[sets.Distribution, sets.Sample],
    r: ScalarList,
    **kwargs,
) -> AnalysisOutput:
    """
    Return the values of the quantities in q calculated incrementally.

    The following quantities can be requested:
        'MMMM': stacked values of M++, M-+, M+-, M-- [1]
        'KKKK': stacked values of K++, K-+, K+-, K-- [nm^n]
        'gggg': stacked values of g++, g-+, g+-, g-- [1]
        'GaGs': stacked values of Ga and Gs [nm^-1]
        'cp': number of dislocations with sense + [1]
        'cm': number of dislocations with sense - [1]
        'dp': density of dislocations with sense + [nm^-n]
        'dm': density of dislocations with sense - [nm^-n]
        'dV': differentials of the neighborhood volume [nm^n]
        'dKKKK': differentials of K++, K-+, K+-, K-- [nm^n]
        'dMMMM': differentials of M++, M-+, M+-, M-- [1]

    Input:
        q (list): name of the quantities to calculate
        o (Distribution|Sample): object to analyze
        r (ScalarList): neighborhood radii in ascending order [nm]
      **r2 (ScalarList): squared neighborhood radii [nm^2]
      **dr (ScalarList): differentials of the neighborhood radii [nm]
      **ec (str): edge consideration (default: 'NEC')

    Output:
        v (AnalysisOutput): values of the quantities in requested order

    Exemple:
        KKKK, GaGs = calculate(['KKKK', 'GaGs'], distribution, r)

    Complexity:
        O( complexity_of(MMMM_Pp_Pm) ) if o is a distribution
        O( complexity_of(MMMM_Pp_Pm) * len(o) ) if o is a sample
    """
    # optional parameters
    r2 = getkwa('r2', kwargs, ScalarList, np.square(r))
    dr = getkwa('dr', kwargs, ScalarList, np.gradient(r))
    ec = getkwa('ec', kwargs, str, 'NEC')
    endkwa(kwargs)
    # create an incremental analysis function to be applied to a distribution
    @beartype
    def calculate_on_distribution(
        d: sets.Distribution,
    ) -> AnalysisOutput:
        """
        Auxiliary function applied to a distribution only.
        """
        s = {} # results temporary storage
        # flags determining what to calculate
        clcgggg = 'gggg' in q or 'dp' in q or 'dm' in q or 'dKKKK' in q
        clcKKKK = clcgggg or 'KKKK' in q or 'cp' in q or 'cm' in q
        clcGaGs = 'GaGs' in q or 'dMMMM' in q
        # calculate
        s['MMMM'], s['cp'], s['cm'] = MMMM_cp_cm(
            d,
            r,
            r2,
            ec,
        )
        if clcKKKK:
            s['KKKK'], s['dp'], s['dm'] = KKKK_dp_dm(
                s['MMMM'],
                s['cp'],
                s['cm'],
                d.v,
            )
        if clcgggg or clcGaGs:
            s['dr'] = np.gradient(r)
        if clcgggg:
            s['gggg'], s['dV'], s['dKKKK'] =  gggg_dV_dKKKK(
                s['KKKK'],
                s['dr'],
                r,
                d.n,
            )
        if clcGaGs:
            s['GaGs'], s['dMMMM'] = GaGs_dMMMM(
                s['MMMM'],
                s['dr'],
                s['cp'],
                s['cm'],
            )
        return tuple([s[k] for k in q])
    # return the result of the function applied to a distribution or a sample
    if isinstance(o, sets.Distribution):
        return calculate_on_distribution(o)
    return o.average(calculate_on_distribution)

@beartype
def plot_KKKK(
    r: ScalarList,
    KKKK: ScalarListList,
    **kwargs,
) -> None:
    """
    Export the figure showing functions K++, K-+, K+-, K--.

    Input:
        r (ScalarList): radius of the neighborhoods [nm]
        KKKK (ScalarListList): stacked K++, K-+, K+-, K-- values [nm^n]
      **expdir (str): export directory (default: '')
      **expfmt (str): export format (default: 'pdf')
      **expstm (str): export stem (default: 'KKKK')
      **figttl (str): figure title (default: "")
      **edgcon (str): edge consideration (default: "")

    Complexity:
        O( len(r) )
    """
    # optional parameters
    expdir = getkwa('expdir', kwargs, str, '')
    expfmt = getkwa('expfmt', kwargs, str, 'pdf')
    expstm = getkwa('expstm', kwargs, str, 'KKKK')
    figttl = getkwa('figttl', kwargs, str, "")
    edgcon = getkwa('edgcon', kwargs, str, "")
    endkwa(kwargs)
    # K of a uniform distribution
    r_compare = np.linspace(r[0], r[-1], 10)
    k_compare = np.pi*r_compare**2
    # fig
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 6))
    fig.subplots_adjust(left=0.06, right=0.98, bottom=0.1)
    fig.suptitle(figttl, fontsize=16)
    # ax1
    ax1.plot(r, KKKK[0], label=r"$K_{++}^{"+edgcon+r"}(r)$")
    ax1.plot(r, KKKK[1], label=r"$K_{-+}^{"+edgcon+r"}(r)$")
    ax1.plot(r_compare, k_compare, "*", label=r"$ \pi r^2 $", color='black')
    ax1.legend()
    ax1.grid()
    ax1.set_xlabel(r"$r \ (nm)$")
    # ax2
    ax2.plot(r, KKKK[3], label=r"$K_{--}^{"+edgcon+r"}(r)$")
    ax2.plot(r, KKKK[2], label=r"$K_{+-}^{"+edgcon+r"}(r)$")
    ax2.plot(r_compare, k_compare, "*", label=r"$ \pi r^2 $", color='black')
    ax2.legend()
    ax2.grid()
    ax2.set_xlabel(r"$r \ (nm)$")
    ax2.set_ylabel(r"$(nm^2)$")
    # export
    plt.savefig(os.path.join(expdir, expstm+"."+expfmt), format=expfmt)
    plt.close('all')

@beartype
def plot_gggg(
    r: ScalarList,
    gggg: ScalarListList,
    **kwargs,
) -> None:
    """
    Export the figure showing functions g++, g-+, g+-, g--.

    Input:
        r (ScalarList): radius of the neighborhoods [nm]
        gggg (ScalarListList): stacked g++, g-+, g+-, g-- values [1]
      **expdir (str): export directory (default: '')
      **expfmt (str): export format (default: 'pdf')
      **expstm (str): export stem (default: 'gggg')
      **figttl (str): figure title (default: "")
      **edgcon (str): edge consideration (default: "")

    Complexity:
        O( len(r) )
    """
    # optional parameters
    expdir = getkwa('expdir', kwargs, str, '')
    expfmt = getkwa('expfmt', kwargs, str, 'pdf')
    expstm = getkwa('expstm', kwargs, str, 'gggg')
    figttl = getkwa('figttl', kwargs, str, "")
    edgcon = getkwa('edgcon', kwargs, str, "")
    endkwa(kwargs)
    # fig
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 6))
    fig.subplots_adjust(left=0.06, right=0.98, bottom=0.1)
    fig.suptitle(figttl, fontsize=16)
    masked = np.ma.masked_invalid(gggg)
    margin = max(np.ptp(masked)*0.05, 0.1)
    ymin = np.min(masked) - margin
    ymax = np.max(masked) + margin
    # ax1
    ax1.plot(r, gggg[0], label=r"$g_{++}^{"+edgcon+r"}(r)$")
    ax1.plot(r, gggg[1], label=r"$g_{-+}^{"+edgcon+r"}(r)$")
    ax1.hlines(1, r[0], r[-1], label=r"$1$", color='black')
    ax1.legend()
    ax1.grid()
    ax1.set_xlabel(r"$r \ (nm)$")
    ax1.set_ylim(ymin, ymax)
    # ax2
    ax2.plot(r, gggg[3], label=r"$g_{--}^{"+edgcon+r"}(r)$")
    ax2.plot(r, gggg[2], label=r"$g_{+-}^{"+edgcon+r"}(r)$")
    ax2.hlines(1, r[0], r[-1], label=r"$1$", color='black')
    ax2.legend()
    ax2.grid()
    ax2.set_xlabel(r"$r \ (nm)$")
    ax2.set_ylim(ymin, ymax)
    # export
    plt.savefig(os.path.join(expdir, expstm+"."+expfmt), format=expfmt)
    plt.close('all')

@beartype
def plot_GaGs(
    r: ScalarList,
    GaGs: ScalarListList,
    **kwargs,
) -> None:
    """
    Export the figure showing functions Ga and Gs.

    Input:
        r (ScalarList): radius of the neighborhoods [nm]
        GaGs (ScalarListList): stacked values of Ga and Gs [m^-1]
      **expdir (str): export directory (default: '')
      **expfmt (str): export format (default: 'pdf')
      **expstm (str): export stem (default: 'GaGs')
      **figttl (str): figure title (default: "")
      **edgcon (str): edge consideration (default: "")

    Complexity:
        O( len(r) )
    """
    # optional parameters
    expdir = getkwa('expdir', kwargs, str, '')
    expfmt = getkwa('expfmt', kwargs, str, 'pdf')
    expstm = getkwa('expstm', kwargs, str, 'GaGs')
    figttl = getkwa('figttl', kwargs, str, "")
    edgcon = getkwa('edgcon', kwargs, str, "")
    endkwa(kwargs)
    # fig
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 6))
    fig.subplots_adjust(left=0.06, right=0.98, bottom=0.1)
    fig.suptitle(figttl, fontsize=16)
    masked = np.ma.masked_invalid(GaGs)
    # ax1
    marg1 = max(np.ptp(masked[0])*0.05, 0.1)
    ax1.plot(r, GaGs[0], label=r"$G_A^{"+edgcon+r"}(r)$")
    ax1.legend()
    ax1.grid()
    ax1.set_xlabel(r"$r \ (nm)$")
    ax1.set_ylim(np.min(masked[0])-marg1, np.max(masked[0])+marg1)
    # ax2
    marg2 = max(np.ptp(masked[1])*0.05, 0.1)
    ax2.plot(r, GaGs[1], label=r"$G_S^{"+edgcon+r"}(r)$")
    ax2.legend()
    ax2.grid()
    ax2.set_xlabel(r"$r \ (nm)$")
    ax2.set_ylim(np.min(masked[1])-marg2, np.max(masked[1])+marg2)
    # export
    plt.savefig(os.path.join(expdir, expstm+"."+expfmt), format=expfmt)
    plt.close('all')

@beartype
def export(
    o: Union[sets.Distribution, sets.Sample],
    **kwargs,
) -> tuple:
    """
    Export a complete analysis of the object o.

    Input:
        o (Distribution|Sample): object to analyze
      **expdir (str): export directory (default: '')
      **expfmt (str): export format (default: 'pdf')
      **expstm (str): export stem (default: o.name(fstm, c='stm'))
      **figttl (str): figure title (default: o.name(fttl, c='ttl'))
      **edgcon (str): edge consideration (default: 'NEC')
      **intrad (ScalarList): interval of radii [nm] (default: ROI size)
      **savtxt (bool): save data to text files (default: False)

    Output:
        r (ScalarList): radius of the neighborhoods [nm]
        KKKK (ScalarListList): stacked K++, K-+, K+-, K-- values [nm^n]
        gggg (ScalarListList): stacked g++, g-+, g+-, g-- values [1]
        GaGs (ScalarListList): stacked Ga and Gs values [m^-1]

    Complexity:
        O( complexity_of(calculate) )
    """
    # optional parameters
    fstm, fttl = 'dmgscS', 'mgscd'
    if isinstance(o, sets.Sample):
        fstm, fttl = "n"+fstm, "n"+fttl
    rmax = o.s
    if o.g == 'circle':
        rmax *= 2
    expstm = getkwa('expstm', kwargs, str, o.name(fstm, c='stm'))
    figttl = getkwa('figttl', kwargs, str, o.name(fttl, c='ttl'))
    edgcon = getkwa('edgcon', kwargs, str, 'NEC')
    intrad = getkwa('intrad', kwargs, ScalarList, np.linspace(0, rmax, 200))
    savtxt = getkwa('savtxt', kwargs, bool, False)
    kwargs['figttl'] = figttl
    kwargs['edgcon'] = edgcon
    # export
    fun = ('KKKK', 'gggg', 'GaGs') # functions to calculate
    KKKK, gggg, GaGs = calculate(fun, o, intrad, ec=edgcon) # function results
    KKKKstm = expstm+"_KKKK_"+edgcon
    ggggstm = expstm+"_gggg_"+edgcon
    GaGsstm = expstm+"_GaGs_"+edgcon
    plot_KKKK(intrad, KKKK, **kwargs, expstm=KKKKstm)
    plot_gggg(intrad, gggg, **kwargs, expstm=ggggstm)
    plot_GaGs(intrad, GaGs, **kwargs, expstm=GaGsstm)
    if savtxt:
        pth = kwargs.get('expdir', '')
        np.savetxt(os.path.join(pth, expstm+"_radii.txt"), intrad)
        np.savetxt(os.path.join(pth, KKKKstm+'.txt'), KKKK)
        np.savetxt(os.path.join(pth, ggggstm+'.txt'), gggg)
        np.savetxt(os.path.join(pth, GaGsstm+'.txt'), GaGs)
    return intrad, KKKK, gggg, GaGs



#!/usr/bin/env python
# coding: utf-8

"""
Tools to standardize notations.

The following abbreviations define the contexts:
    stm: the string is used in the stem of a file name
    ttl: the string is a title and contains LaTeX code
    csl: the string is displayed in a console
"""

from . import *

@beartype
def number(
    x: Scalar,
    c: str = 'csl',
    w: int = 4,
) -> str:
    """
    Return a notation of the number x in context c.

    Input:
        x (Scalar): number
        c (str): context
        w (int): width of the output string

    Output:
        s (str): notation of x
    """
    if x<10**w and x>=1 or x == 0: # diplay x with fixed comma
        s = str(round(x))
        if c == 'stm':
            s = s.zfill(w) # fill with zeros
        elif c == 'csl':
            s = format(s, '>'+str(w)) # fill with spaces
    elif c == 'ttl': # LaTeX scientific notation
        s = format(x, '1.'+str(max(w-6, 0))+'e')
        m, e = s.split("e") # separate mantissa and exponent
        s = "$ "+m+" \\times 10^{"+str(int(e))+"} $"
    elif c in ('stm', 'csl'): # scientific notation with 'e'
        s = format(x, '1.0e').replace("+", "")
        s = s.replace("e0", "e").replace("e-0", "e-")
        d = w-len(s) # number of missing characters to reach the ideal width
        if d>0:
            if c=='stm' or d==1:
                if "e-" in s:
                    s = s.replace("e-", "e-0")
                else:
                    s = s.replace("e", "e0")
            else:
                fmt = '1.'+str(d-1)+'e'
                s = format(x, fmt)
                s = s.replace("+", "").replace("e0", "e").replace("e-0", "e-")
    else:
        raise ValueError("unknown context: "+str(c))
    return s

@beartype
def unit(
    x: str,
    c: str = 'csl',
) -> str:
    """
    Return a notation of the unit x in context c.

    Input:
        x (str): unit
        c (str): context

    Output:
        s (str): notation of x
    """
    if c == 'ttl':
        return "$ \mathrm{"+x+"} $"
    else:
        return x.replace("{", "").replace("}", "").replace("^", "")

@beartype
def quantity(
    v: Scalar,
    u: str,
    c: str = 'csl',
    w: int = 4,
) -> str:
    """
    Return the notation of value v and unit u in context c.

    Input:
        v (Scalar): value
        u (str): unit
        c (str): context
        w (int): width of the value

    Output:
        s (str): notation of u and v as a physical quantity
    """
    val = number(v, c, w)
    uni = unit(u, c)
    if c == 'ttl':
        val = val.replace(r"$", "").strip()
        uni = uni.replace(r"$", "").strip()
        return "$ "+val+" "+uni+" $"
    else:
        return val+uni

@beartype
def equality(
    a: str,
    b: str,
    c: str = 'csl',
) -> str:
    """
    Return the notation of a=b in context c.

    Input:
        a (str): left term of equality
        b (str): right term of equality
        c (str): context

    Output:
        s (str): notation of the equality
    """
    if c == 'ttl':
        a = a.replace(r"$", "").strip()
        b = b.replace(r"$", "").strip()
        return r"$ "+a+" = "+b+" $"
    else:
        a = a.replace("\\", "")
        if c == 'csl':
            return a+"="+b
        elif c == 'stm':
            return a+b

@beartype
def parameters(
    r: dict,
    c: str = 'csl',
    s: bool = True,
) -> str:
    """
    Return the notation descibing the model parameters set r.

    Input:
        r (dict): model parameters
        c (str): context
        s (bool): display the seed

    Output:
        n (str): model parameters description
    """
    # user-defined name for the set of parameters
    if 'name' in r:
        return r['name']
    n = ""
    # version
    if 'v' in r:
        n += "-"+r['v']
    # parameter list
    p = []
    if 'd' in r:  # density
        p.append(equality('d', quantity(r['d'], "nm^{-2}", c), c))
    if 's' in r: # subarea/cell side
        p.append(equality("s", number(r['s'], c, 4)+unit("nm", c), c))
    if 'f' in r: # filling
        p.append(equality("f", str(r['f']), c))
    if 't' in r: # wall thickness
        p.append(equality("t", number(r['t'], c, 3)+unit("nm", c), c))
    if 'l' in r: # dipole length
        p.append(equality("l", number(r['l'], c, 3)+unit("nm", c), c))
    # concatenate
    if len(p) > 0:
        if c == 'stm':
            n += "_"+"_".join(p)
        elif c == 'ttl':
            p = [q.replace("$", "").strip() for q in p]
            n += r" $ \left( "+", \ ".join(p)+r" \right) $"
        elif c == 'csl':
            n += " ("+" ".join(p)+")"
        else:
            raise ValueError("unknown context "+str(c))
    return n

@beartype
def fmt(
    f: str,
    v: dict,
    c: str = 'csl',
) -> str:
    """
    Return the elements of v selected and ordered by f.

    Input:
        f (str): format
        v (dict): dictionary containing the parameters
        c (str): context

    Output:
        s (str): elements of v selected and ordered by f
    """
    if c == 'stm':
        sep = '_'
    else:
        sep = " "
    return sep.join([v[n] for n in f if not v[n] is None])



#!/usr/bin/env python
# coding: utf-8

"""
Tools for managing the possible geometries of the region of interest.
"""

from . import *

@beartype
def nvolume(
    g: str,
    s: Scalar
) -> tuple:
    """
    Return the dimension of space n and the n-volume of the shape.

    Input:
        g (str): geometry of the region of interest
        s (Scalar): size of the region of interest [nm]

    Output:
        n (int): dimension of space
        v (Scalar): n-volume of the shape [nm^n]
    """
    if g == 'circle':
        return 2, np.pi*s**2
    elif g == 'square':
        return 2, s**2
    else:
        raise ValueError("unknown geometry: "+str(g))

@beartype
def mask(
    g: str,
    s: Scalar,
    p: VectorList,
) -> np.ndarray:
    """
    Return the mask filtering positions inside the region of interest.

    Input:
        g (str): geometry of the region of interest
        s (Scalar): size of the region of interest [nm]
        p (VectorList): positions of the dislocations [nm]

    Output:
        m (np.ndarray): mask of positions within the region of interest
    """
    if g == 'circle':
        return np.sum(np.square(p), axis=1) < s**2
    elif g == 'square':
        return (p[:,0]<s) & (p[:,1]<s) & (p[:,0]>0) & (p[:,1]>0)
    else:
        raise ValueError("unknown geometry: "+str(g))



#!/usr/bin/env python
# coding: utf-8

"""
Tools for the generation of special conditions at the boundaries.
"""

from . import *

@beartype
def image_positions(
    s: Scalar,
    p: VectorList,
) -> VectorList:
    """
    Return the image screw positions for a circle of radius s.

    Adding the image dislocations allows to simulate correctly the
    complete diffraction profile (and not only for the asymptotic
    values of the Fourier variable). However, the formula used here
    only applies to dislocations of type 'screw'.

    Input:
        s (Scalar): radius of the region of interest [nm]
        p (VectorList): dislocation positions [nm]

    Output:
        cp (VectorList): image dislocation positions [nm]

    Input example:
        s = 2
        p = np.array([[1, 0], [-1, 0], [0, 0]])

    Output example:
        cp = np.array([[4, 0], [-4, 0]])

    Complexity:
        O( len(p) )
    """
    n = np.linalg.norm(p, axis=1) # distance to the origin
    m = n != 0 # mask to avoid division by zero
    v = p[m] # points different from (0, 0)
    t = np.arctan2(v[:,1], v[:,0]) # angles of the image dislocations
    r = s**2/n[m] # radii of the image dislocations
    cp = np.stack((r*np.cos(t), r*np.sin(t)), axis=1) # image positions
    return cp

@beartype
def replication_displacements(
    r: int,
    s: Scalar = 1,
) -> VectorList:
    """
    Return the displacements for the replication of a square of side s.

    Input:
        r (int): rank of replications of the region of interest
        s (Scalar): side of the region of interest

    Output:
        u (VectorList): displacements for the replications

    Input example:
        r = 1
        s = 1

    Output example:
        u = np.array([
            [-2  2],
            [ 2  2],
            [ 2 -2],
            [-2 -2],
            [-2  0],
            [ 0  2],
            [ 2  0],
            [ 0 -2],
        ])

    Complexity:
        O( r^2 )
    """
    u = [] # list containing displacements for the replications
    for i in range(1, r+1):
        for j in range(2*i):
            for k in (1, -1):
                u.append(np.array((-i*k, (i-j)*k)))
                u.append(np.array(((i-j)*k,  i*k)))
    return s*np.array(u)



#!/usr/bin/env python
# coding: utf-8

"""
Tools for calculating the overlapping of two geometric objects.
"""

import scipy.integrate
from . import *

@beartype
def circular_segment(
    r: Union[Scalar, ScalarList],
    d: Union[Scalar, ScalarList],
) -> Union[Scalar, ScalarList]:
    """
    Return the circular segment area.

    Input:
        r (Scalar|ScalarList): circle radius/ii
        d (Scalar|ScalarList): distance(s) of the chord from the center

    Output:
        o (Scalar|ScalarList): circular segment area
    """
    return r**2*np.arccos(d/r) - d*np.sqrt(r**2-d**2)

@beartype
def circle_circle(
    rA: Union[Scalar, ScalarList],
    rB: Union[Scalar, ScalarList],
    d: Union[Scalar, ScalarList],
    r2A: Union[Scalar, ScalarList],
    r2B: Union[Scalar, ScalarList],
    d2: Union[Scalar, ScalarList],
) -> Union[Scalar, ScalarList]:
    """
    Return the overlapping area of two circles.

    All input parameters can be either an array or a scalar. If one of
    them is an array, the result will be an array of the same size.

    Input:
        rA (Scalar|ScalarList): radius/ii of the circle A
        rB (Scalar|ScalarList): radius/ii of the circle B
        d (Scalar|ScalarList): distance(s) between the circle centers
        r2A (Scalar|ScalarList): squared radius/ii of the circle A
        r2B (Scalar|ScalarList): squared radius/ii of the circle B
        d2 (Scalar|ScalarList): squared distance(s) between the centers

    Output:
        o (Scalar|ScalarList): overlapping area/s

    Complexity:
        O( max(rA.size, rB.size, d.size) )
    """
    m0 = rA + rB <= d # mask: zero intersection
    mA = d + rA <= rB # mask: A is inside B
    mB = d + rB <= rA # mask: B is inside A
    m = np.logical_not(m0) & np.logical_not(mA) & np.logical_not(mB)
    # m is true for non-trivial cases
    o = np.subtract(
        np.add(
            np.multiply(
                r2A,
                np.arccos(
                    np.divide(
                        np.add(d2, np.subtract(r2A, r2B, where=m), where=m),
                        np.multiply(np.multiply(2, d, where=m), rA, where=m),
                        where=m),
                    where=m),
                where=m),
            np.multiply(
                r2B,
                np.arccos(
                    np.divide(
                        np.add(d2, np.subtract(r2B, r2A, where=m), where=m),
                        np.multiply(np.multiply(2, d, where=m), rB, where=m),
                        where=m),
                    where=m),
                where=m),
            where=m),
        np.divide(
            np.sqrt(
                np.multiply(
                    np.multiply(
                        np.subtract(np.add(rA, rB, where=m), d, where=m),
                        np.add(np.add(rA, rB, where=m), d, where=m),
                        where=m),
                    np.multiply(
                        np.add(np.subtract(rA, rB, where=m), d, where=m),
                        np.add(np.subtract(rB, rA, where=m), d, where=m),
                        where=m),
                    where=m),
                where=m),
            2,
            where=m),
        where=m)
    o = np.where(mA, np.pi*r2A, o)
    o = np.where(mB, np.pi*r2B, o)
    o = np.where(m0, 0, o)
    return o

@beartype
def circle_square(
    x: Union[Scalar, ScalarList],
    y: Union[Scalar, ScalarList],
    r: Union[Scalar, ScalarList],
    r2: Union[Scalar, ScalarList],
    s: Union[Scalar, ScalarList],
) -> Union[Scalar, ScalarList]:
    """
    Return the overlapping area of a circle and a square.

    All input parameters can be either an array or a scalar. If one of
    them is an array, the result will be an array of the same size.

    Input:
        x (Scalar|ScalarList): x coordinate(s) of the circle center
        y (Scalar|ScalarList): y coordinate(s) of the circle center
        r (Scalar|ScalarList): circle radius/ii
        r2 (Scalar|ScalarList): squared circle radius/ii
        s (Scalar|ScalarList): square side(s)

    Output:
        o (Scalar|ScalarList): overlapping area/s

    Complexity:
        O( max(x.size, r.size, s.size) )
    """
    e = [] # list of areas of the circle outside the square for each frame
    dA = s - x # width of the right quadrant
    dB = s - y # height of the upper quadrant
    dC = x # width of the left quadrant
    dD = y # height of the lower quadrant
    # loop on the four quadrants and add their contributions to E
    for d1, d2 in [[dA, dB], [dB, dC], [dC, dD], [dD, dA]]:
        mq = d1**2 + d2**2 > r2
        m1 = mq & (d1 < r)
        m2 = mq & (d2 < r)
        mr = np.logical_not(mq)
        # area of the circle outside the square in the current quadrant
        eq = np.where(mr, np.subtract(np.pi*r2/4, d1*d2, where=mr), 0)
        for a in [[m1, d1], [m2, d2]]:
            np.add(
                np.divide(
                    np.subtract(
                        np.multiply(
                            r2,
                            np.arccos(
                                np.divide(a[1], r, where=a[0]),
                                where=a[0]),
                            where=a[0]),
                        np.multiply(
                            a[1],
                            np.sqrt(
                                np.subtract(
                                    r2,
                                    np.multiply(a[1], a[1], where=a[0]),
                                    where=a[0]),
                                where=a[0]),
                            where=a[0]),
                        where=a[0]),
                    2,
                    where=a[0]),
                eq,
                eq,
                where=a[0])
        e.append(eq)
    return np.pi*r2 - sum(e)

@np.vectorize
@beartype
def mean_circle_circle_analytic(
    r: Scalar,
    R: Scalar,
) -> Scalar:
    """
    Return the mean overlapping area of two circles.

    All input parameters can be either an array or a scalar. If one of
    them is an array, the result will be an array of the same size.

    Input:
        r (Scalar|ScalarList): circle 1 radius/ii
        R (Scalar|ScalarList): circle 2 radius/ii

    Output:
        o (Scalar|ScalarList): mean overlapping area/s

    Complexity:
        O( r.size )
    """
    r2 = r**2
    R2 = R**2
    f = lambda x: circle_circle(r, R, R*np.sqrt(x), r2, R2, R2*x)
    return scipy.integrate.quad(f, 0, 1)[0]

@np.vectorize
@beartype
def mean_circle_square_analytic(
    r: Scalar,
    s: Scalar,
) -> Scalar:
    """
    Return the mean overlapping area of a circle and a square.

    All input parameters can be either an array or a scalar. If one of
    them is an array, the result will be an array of the same size.

    Input:
        r (Scalar|ScalarList): circle radius/ii
        s (Scalar|ScalarList): square side(s)

    Output:
        o (Scalar|ScalarList): mean overlapping area/s

    Complexity:
        O( r.size )
    """
    f12 = lambda x, phi: circular_segment(r, x*np.cos(phi))*x/(2*s**2)
    f3a = lambda x, phi: (np.pi*r**2/4 - x*np.cos(phi)*x*np.sin(phi))*x/(s**2)
    f3b = lambda x, y: (np.pi*r**2/4 - x*y)/s**2
    phi1 = np.arccos(s/max(r, s))
    phi2 = np.arctan(s/min(r, s))
    phi3 = np.arcsin(s/max(r, s))
    x1 = lambda phi: min(r, s)/np.cos(phi)
    x2 = lambda phi: s/np.sin(phi)
    x3 = lambda y: np.tan(phi1)*y
    if r == 0:
        return 0
    elif r <= np.sqrt(2)*s:
        E12 = (scipy.integrate.dblquad(f12, phi1, phi2, r, x1)[0]
            + scipy.integrate.dblquad(f12, phi2, phi3, r, x2)[0])
        E3 = (scipy.integrate.dblquad(f3a, phi1, phi3, 0, r)[0]
            + 2*scipy.integrate.dblquad(f3b, 0, s, 0, x3)[0])
        return np.pi*r**2 - 4*(2*E12+E3)
    else:
        return s**2

@beartype
def mean_circle_circle_simulation(
    r: Union[Scalar, ScalarList],
    R: Union[Scalar, ScalarList],
    n: int = 1000000,
    G: np.random._generator.Generator = np.random.default_rng(0),
) -> Union[Scalar, ScalarList]:
    """
    Return the mean overlapping area of two circles.

    Input parameters r and R can be either an array or a scalar. If one
    of them is an array, the result will be an array of the same size.

    Input:
        r (Scalar|ScalarList): circle 1 radius/ii
        R (Scalar|ScalarList): circle 2 radius/ii
        n (int): number of tested positions
        G (np.random._generator.Generator): random number generator

    Output:
        o (Scalar|ScalarList): mean overlapping area/s

    Complexity:
        O( r.size )
    """
    d = R*np.sqrt(G.random(n))
    d2 = d**2
    m = lambda r, R: circle_circle(r, R, d, r**2, R**2, d2).mean()
    return np.vectorize(m)(r, R)

@beartype
def mean_circle_square_simulation(
    r: Union[Scalar, ScalarList],
    s: Union[Scalar, ScalarList],
    n: int = 1000000,
    G: np.random._generator.Generator = np.random.default_rng(0),
) -> Union[Scalar, ScalarList]:
    """
    Return the mean overlapping area of a circle and a square.

    All input parameters can be either an array or a scalar. If one of
    them is an array, the result will be an array of the same size.

    Input:
        r (Scalar|ScalarList): circle radius/ii
        s (Scalar|ScalarList): square side(s)
        n (int): number of tested positions
        G (np.random._generator.Generator): random number generator

    Output:
        o (Scalar|ScalarList): mean overlapping area/s

    Complexity:
        O( r.size )
    """
    x = G.random(n)*s
    y = G.random(n)*s
    m = lambda ri, si: circle_square(x, y, ri, ri**2, si).mean()
    return np.vectorize(m)(r, s)



#!/usr/bin/env python
# coding: utf-8

"""
Tools for parallelizing the spatial analysis of distributions.
"""

from mpi4py import MPI
name = MPI.Get_processor_name()
comm = MPI.COMM_WORLD
rank = comm.Get_rank() # number of the processor executing this script
size = comm.Get_size() # number of processors
root = 0 # master processor
from . import *
from . import sets
from . import analyze

@beartype
def average_on_cores(
    w: AnalysisOutput,
    b: bool = False
) -> Optional[AnalysisOutput]:
    """
    Return the average value of w ​​over the cores.

    When b is True, all cores get the average value. When b is False,
    the workers get None and the root gets the average value.

    Input:
        w (AnalysisOutput): worker value
        b (bool): broadcast the result to all cores

    Output:
        m (AnalysisOutput): averaged value of w over the cores

    Input example:
        On master core:
            w = 1
            b = False
        On worker core:
            w = 0
            b = False

    Output example:
        On master core:
            m = 0.5
        On worker core:
            m = None
    """
    if not isinstance(w, np.ndarray):
        w = np.array(w)
    if rank == root: # if the script is executed by the master
        m = np.zeros_like(w) # prepare the buffer for data reception
    else: # if the script is executed by a worker
        m = None
    comm.Reduce([w, MPI.DOUBLE], [m, MPI.DOUBLE], op=MPI.SUM, root=root)
    if rank == root:
        m = m/size # average the value
    if b:
        m = comm.bcast(m, root=root) # broadcast to the workers
    return m

@beartype
def export(
    o: Union[sets.Distribution, sets.Sample],
    **kwargs,
) -> Optional[tuple]:
    """
    Export a complete pooled analysis of the object o of each core.

    Function similar to the export function of the analyze module.

    Input:
        o (Distribution|Sample): object to analyze on the core
      **expdir (str): export directory (default: '')
      **expfmt (str): export format (default: 'pdf')
      **expstm (str): export stem (default: o.name(fstm, c='stm'))
      **figttl (str): figure title (default: o.name(fttl, c='ttl'))
      **edgcon (str): edge consideration (default: 'NEC')
      **intrad (ScalarList): interval of radii [nm] (default: ROI size)
      **savtxt (bool): save data to text files (default: False)

    Output:
        r (ScalarList): radius of the neighborhoods [nm]
        KKKK (ScalarListList): stacked K++, K-+, K+-, K-- values [nm^n]
        gggg (ScalarListList): stacked g++, g-+, g+-, g-- values [1]
        GaGs (ScalarListList): stacked Ga and Gs values [m^-1]
    """
    if rank==root and not o.S is None:
        msg = ("chosen random seed detected on "+o.name('stm')+" "
            + "(The use of a seed is not recommended. In order for "
            + "the distributions or samples not to be identical from "
            + "one core to another, the seed must not be identical "
            + "from one core to another.")
        warnings.warn(msg, Warning)
    rmax = o.s
    if o.g == 'circle':
        rmax *= 2
    edgcon = getkwa('edgcon', kwargs, str, 'NEC')
    intrad = getkwa('intrad', kwargs, ScalarList, np.linspace(0, rmax, 200))
    i = average_on_cores(o.i, True) # averaged inter dislocation distance
    fun = ('KKKK', 'gggg', 'GaGs') # functions to calculate
    worker = analyze.calculate(fun, o, intrad, ec=edgcon) # function results
    master = [average_on_cores(worker[i]) for i in range(len(worker))]
    if rank == root:
        if isinstance(o, sets.Distribution):
            c = str(size) # number of distributions analyzed
        else:
            c = str(len(o)*size) # number of distributions analyzed
        # optional parameters
        expstm = getkwa('expstm', kwargs, str, c+"_"+o.name('dmgsS', c='stm'))
        figttl = getkwa('title', kwargs, str, c+" "+o.name('mgsd', c='ttl'))
        savtxt = getkwa('savtxt', kwargs, bool, False)
        kwargs['figttl'] = figttl
        kwargs['edgcon'] = edgcon
        # export
        KKKK, gggg, GaGs = master
        KKKKstm = expstm+"_KKKK_"+edgcon
        ggggstm = expstm+"_gggg_"+edgcon
        GaGsstm = expstm+"_GaGs_"+edgcon
        analyze.plot_KKKK(intrad, KKKK, **kwargs, expstm=KKKKstm)
        analyze.plot_gggg(intrad, gggg, **kwargs, expstm=ggggstm)
        analyze.plot_GaGs(intrad, GaGs, **kwargs, expstm=GaGsstm)
        if savtxt:
            pth = kwargs.get('expdir', '')
            np.savetxt(os.path.join(pth, expstm+"_radii.txt"), intrad)
            np.savetxt(os.path.join(pth, KKKKstm+'.txt'), KKKK)
            np.savetxt(os.path.join(pth, ggggstm+'.txt'), gggg)
            np.savetxt(os.path.join(pth, GaGsstm+'.txt'), GaGs)
        return intrad, KKKK, gggg, GaGs
